# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import xtrack as xt
import xfields as xf
import xobjects as xo
import xcoll as xc


# ===============================================================================================
# IMPORTANT
# ===============================================================================================
# Only make changes to this file just before a minor version bump (need a separate commit though)
# to avoid having multiple xboinc versions with out-of-sync executables.
# ===============================================================================================

def get_default_tracker():
    """
    Returns a default tracker object.
    """

    # Dummy line containing all supported element types
    default_line = xt.Line(elements=[
        xt.Marker(),
        xt.Drift(),
        xt.Bend(length=1.0),
        xt.Multipole(),
        xt.CombinedFunctionMagnet(length=1.0),
        xt.Quadrupole(length=1.0),
        xt.Sextupole(length=1.0),
        xt.SimpleThinBend(),
        xt.SimpleThinQuadrupole(),
        xt.ReferenceEnergyIncrease(),
        xt.Cavity(),
        xt.Solenoid(length=1.0),
        xt.XYShift(),
        xt.Elens(),
        xt.NonLinearLens(),
        xt.Wire(),
        xt.SRotation(),
        xt.XRotation(),
        xt.YRotation(),
        xt.ZetaShift(),
        xt.RFMultipole(knl=[0], pn=[0]),
#        xt.Fringe(),
#        xt.Wedge(),
        xt.DipoleEdge(),
        xt.Exciter(nsamples=1),
        xt.LineSegmentMap(),
        xt.FirstOrderTaylorMap(),
        xf.BeamBeamBiGaussian2D(
            other_beam_Sigma_11=1.,
            other_beam_Sigma_33=1.,
            other_beam_num_particles=0.,
            other_beam_q0=1.,
            other_beam_beta0=1.,
        ),
#         # Doesn't work because issue with definition of atomicAdd when generating executable
#         xf.BeamBeamBiGaussian3D(
#             slices_other_beam_zeta_center=[0],
#             slices_other_beam_num_particles=[0],
#             phi=0.,
#             alpha=0,
#             other_beam_q0=1.,
#             slices_other_beam_Sigma_11=[1],
#             slices_other_beam_Sigma_12=[0],
#             slices_other_beam_Sigma_22=[0],
#             slices_other_beam_Sigma_33=[1],
#             slices_other_beam_Sigma_34=[0],
#             slices_other_beam_Sigma_44=[0],
#         ),
#         # Doesn't work because fieldmap in different buffer
#         xf.ElectronCloud(fieldmap=xf.TriCubicInterpolatedFieldMap(
#             x_range=(0.,1.), nx=10,
#             y_range=(0.,1.), ny=10,
#             z_range=(0.,1.), nz=10
#         )),
#         # Doesn't work because issue with definition of atomicAdd when generating executable
#         xf.ElectronLensInterpolated(
#             x_range=(0.,1.), nx=10,
#             y_range=(0.,1.), ny=10
#         ),
#         # Doesn't work because issue with definition of atomicAdd when generating executable
#         xf.SpaceCharge3D(
#             x_range=(0.,1.), nx=10,
#             y_range=(0.,1.), ny=10,
#             z_range=(0.,1.), nz=10
#         ),
        xc.BlackAbsorber(length=1),
        xc.EverestCollimator(length=1, material=xc.materials.Silicon),
        xc.EverestCrystal(length=1, material=xc.materials.SiliconCrystal),
        xt.LimitRect(),
        xt.LimitRacetrack(),
        xt.LimitEllipse(),
        xt.LimitPolygon(x_vertices=[0.,1.], y_vertices=[0.,1.]),
        xt.LimitRectEllipse(),
        xt.LongitudinalLimitRect()
    ])

    _context = xo.ContextCpu()
    default_line.build_tracker(_context=_context, compile=True,
                               use_prebuilt_kernels=False)
    default_config_hash = default_line.tracker._hashable_config()
    return default_line.tracker, default_config_hash



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import json
import tarfile
from pathlib import Path
import tempfile

from .user import get_domain, get_folder
from .server.eos import mv_from_eos, mv_to_eos, xrdcp_installed
from .server.tools import timestamp
from .simulation_io import SimConfig, app_version


eosdir  = '/eos/user/d/ddicroce/xboinc/'


class SubmitJobs:

    def __init__(self, user, study):
        if '__' in study:
            raise ValueError("The character sequence '__' is not allowed in 'study'!")
        self._username = user
        self._domain = get_domain(user)
        if self._domain=='eos' and not xrdcp_installed():
            raise ValueError("Error: xrdcp is not installed on your system. Cannot submit.")
        self._target = get_folder(user)
        self._studyname = study
        self._submitfile = f"{self._studyname}__{timestamp()}.tar.gz"
        self._json_files = []
        self._bin_files = []
        self._temp    = tempfile.TemporaryDirectory()
        self._tempdir = Path(temp.name).resolve()
        self._tempdir.mkdir(parents=True, exist_ok=True)

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        self._submit()
        self._temp.cleanup()


    def add(self, *, job_name, num_turns, line, particles, checkpoint_every=-1, **kwargs):
        if '__' in job_name:
            raise ValueError("The character sequence '__' is not allowed in 'job_name'!")
        filename = f"{self._username}__{timestamp(ms=True)}"
        json_file = Path(self._tempdir, f"{filename}.json")
        bin_file  = Path(self._tempdir, f"{filename}.bin")
        kwargs['xboinc_ver'] = app_version
        json_dict = {
            'study': self._studyname,
            'user': self._username,
            'job_name': job_name,
            **kwargs
        }
        with json_file.open('w') as fid:
            json.dump(json_dict, fid)
        SimConfig.build(filename=bin_file, num_turns=num_turns, line=line,
                         particles=particles, checkpoint_every=checkpoint_every)
        self._json_files += [json_file]
        self._bin_files  += [bin_file]


    def _submit(self):
        with tarfile.open(self._tempdir / self._submitfile, "w:gz") as tar:
            for thisfile in self._json_files + self._bin_files:
                tar.add(thisfile, arcname=thisfile.name)
        if self._domain == 'eos':
            mv_to_eos(self._tempdir / self._submitfile, self._target)
        elif self._domain == 'afs':
            mv_to_afs(self._tempdir / self._submitfile, self._target)
        else:
            raise ValueError(f"Wrong domain {self._domain} for user {self._username}!")
        # TODO: check that tar contains all files
        for thisfile in self._json_files + self._bin_files:
            thisfile.unlink()




# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from .general import _pkg_root, __version__, __xsuite__versions__

from .register import register
from .submit import SubmitJobs

from .simulation_io import SimState, SimConfig, app_version, app_version_int
from .executable import generate_executable_source, generate_executable


# Check that the active environment has the correct pinned versions
import sys
error = ""
for mod in __xsuite__versions__.keys():
    __import__(mod)
    version = sys.modules[mod].__version__
    if version != __xsuite__versions__[mod]:
        error += f"This xboinc version ({__version__}) needs "
        error += f"{mod} {__xsuite__versions__[mod]}, "
        error +=  f"but active version is {version}!\n"
if error != "":
    raise ValueError(error)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import xobjects as xo
import xpart as xp

from ..general import __version__


# XXX.YYY.ZZZ to XXXYYY  (ignore patch)
def _version_to_int(version):
    vers = version.split('.')
    return int(vers[0])*1000 + int(vers[1])

# XXXYYY to XXX.YYY
def _int_to_version(verint):
    XXX = int(verint / 1000)
    YYY = int(verint - XXX*1000)
    return f"{XXX}.{YYY}"


app_version = '.'.join(__version__.split('.')[:2])
app_version_int = _version_to_int(__version__)


# This class overloads the first field from the SimState,
# in order to read the correct version from the binary
class SimVersion(xo.HybridClass):
    _xofields = {
        'xboinc_version': xo.Int64,    # version XXX.YYY as int
    }

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            kwargs['xboinc_version'] = _version_to_int(__version__)
        super().__init__(**kwargs)

    def assert_version(self):
        if app_version_int != self.xboinc_version:
            error += f"Incompatible xboinc version! Output file needs "
            error +  f"{_int_to_version(self.xboinc_version)}, "
            error +  f"but current version is {__version__}.\n"
            raise ValueError(error)



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import xobjects as xo
from ..simulation_io import SimState, SimConfig, SimVersion, app_version, get_default_tracker
from ..general import _pkg_root, __version__

from pathlib import Path
import shutil
import subprocess

# ===============================================================================================
# IMPORTANT
# ===============================================================================================
# Only make changes to this file just before a minor version bump (need a separate commit though)
# to avoid having multiple xboinc versions with out-of-sync executables.
# ===============================================================================================

insert_in_all_files = """
#include <stdio.h>
#ifndef NULL
    #define NULL 0
#endif
"""

def generate_executable_source(write_source_files=True,
                                       _context=None):
    assert _context is None
    assert write_source_files

    sim_config_sources = [
        insert_in_all_files,
        xo.specialize_source(SimVersion._XoStruct._gen_c_api().source,
                                    specialize_for='cpu_serial'),
        xo.specialize_source(SimState._XoStruct._gen_c_api().source,
                                    specialize_for='cpu_serial'),
        xo.specialize_source(SimConfig._gen_c_api().source,
                                    specialize_for='cpu_serial'),
    ]

    sim_config_h = '\n'.join(sim_config_sources)

    default_tracker, default_config_hash = get_default_tracker()
    track_kernel = default_tracker.track_kernel[default_config_hash]
    xtrack_tracker_h = (
        insert_in_all_files + track_kernel.specialized_source)

    with open(_pkg_root.joinpath('executable/main.c'), 'r') as fid:
        main_c = insert_in_all_files + fid.read()

    dct_sources = {
        'sim_config.h': sim_config_h,
        'xtrack_tracker.h': xtrack_tracker_h,
        'main.c': main_c,
    }

    for nn, vv in dct_sources.items():
        file = Path.cwd() / nn
        with file.open('w') as fid:
            fid.write(vv)

    return dct_sources


def generate_executable(keep_source=False):
    main    = Path.cwd() / "main.c"
    config  = Path.cwd() / "sim_config.h"
    tracker = Path.cwd() / "xtrack_tracker.h"
    if not main.exists() or not config.exists() or not tracker.exists():
        source_files = generate_executable_source()

    if shutil.which("gcc") is not None:
        compiler = "gcc"
    elif shutil.which("clang") is not None:
        compiler = "clang"
    else:
        raise RuntimeError("Neither clang or gcc are found. Install a C compiler.")

    tag = f'_{app_version}'
    cmd = subprocess.run(['uname', '-ms'], stdout=subprocess.PIPE)
    if cmd.returncode == 0:
        tag += '-' + cmd.stdout.decode('UTF-8').strip().lower().replace(' ','-')
    cmd = subprocess.run([compiler, 'main.c', '-o', f'xboinc{tag}', '-lm'],
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if cmd.returncode != 0:
        stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
        raise RuntimeError(f"Compilation failed. Stderr:\n {stderr}")

    if keep_source:
        main.rename(main.parent / f'{main.stem}{tag}{main.suffix}')
        config.rename(config.parent / f'{config.stem}{tag}{config.suffix}')
        tracker.rename(tracker.parent / f'{tracker.stem}{tag}{tracker.suffix}')
    else:
        main.unlink()
        config.unlink()
        tracker.unlink()



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from .default_tracker import get_default_tracker
from .output import SimState
from .input import SimConfig
from .version import SimVersion, app_version, app_version_int


# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from .generate_executable_source import generate_executable_source, generate_executable



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import numpy as np
from pathlib import Path

import xobjects as xo
import xpart as xp
import xtrack as xt

from .default_tracker import get_default_tracker
from .output import SimState

_default_tracker, _default_config_hash = get_default_tracker()


class SimConfig(xo.Struct):
    tracker_data     = _default_tracker._tracker_data_cache[_default_config_hash]
    line_metadata    = xo.Ref(tracker_data._element_ref_data.__class__)
    num_turns        = xo.Int64
    num_elements     = xo.Int64
    checkpoint_every = xo.Int64
    sim_state        = xo.Ref(SimState)

    @classmethod
    def build(cls, *, num_turns, line, particles, checkpoint_every=-1, filename=None):
        # Assemble data structure
        simbuf = xo.ContextCpu().new_buffer()
        sim_config = cls(_buffer=simbuf)
        default_tracker = _default_tracker
        line.build_tracker(_context=simbuf.context, _buffer=simbuf,
                            track_kernel=default_tracker.track_kernel)
        if _default_config_hash not in line.tracker._tracker_data_cache:
            raise RuntimeError('Tracker data for default config not found')
        tracker_data = line.tracker._tracker_data_cache[_default_config_hash]
        sim_state = SimState(particles=particles, i_turn=0, _buffer=simbuf)
        sim_config.line_metadata = tracker_data._element_ref_data
        sim_config.num_turns = num_turns
        sim_config.num_elements = len(line.element_names)
        sim_config.checkpoint_every = checkpoint_every
        sim_config.sim_state = sim_state._xobject
        # TODO: can we set size at construction?
        sim_state.size = sim_state._xobject._size # store size of sim_state
        assert sim_config._offset == 0
        assert sim_config._fields[0].offset == 0
        # Write sim buffer to file
        if filename is not None:
            filename = Path(filename).expanduser().resolve()
            with filename.open('wb') as fid:
                fid.write(simbuf.buffer.tobytes())
        return sim_config



# copyright ############################### #
# This file is part of the Xboinc Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from pathlib import Path

eosdir  = Path('/','eos','user','s','sixtadm','spooldirs','xtrack') # EOS spooldir for server
dropdir = eosdir / 'drop'                                           # only location where users have write access (i.e. to register)
